# Uses dictionary learning (sparse coding) as a preprocessing step,
# i.e., each cell image is split up as lots of patches, and each patch
# is a linear combination of elements in the "dictionary". In other
# works, where there was a pixel, there is now a vector of weights
# with one element for each word in the dictionary.

import os
import cPickle
import pylab as pb
import numpy as np
import GPy
import Image
pb.ion()

# Based on the example in:
# http://scikit-learn.org/stable/modules/decomposition.html#dictionary-learning
from time import time

from scipy.misc import lena # sample image

from sklearn.decomposition import MiniBatchDictionaryLearning
from sklearn.feature_extraction.image import extract_patches_2d
from sklearn.feature_extraction.image import reconstruct_from_patches_2d


###############################################################################
print('Load images and extract patches, or load pickled data')

print('TO DO: build dictionary combining positive and negative samples!!!')

image_type = 'RotatedPositives2' # 'Positives' 
im_dir = os.path.join(os.path.expanduser('~'), 'Datasets', 'AMIDA2013', 'Patches', image_type)
precompt_dir = os.path.join(os.path.expanduser('~'), 'PrecomputedLaplace', 'AMIDA2013', image_type, 'DictionaryLearning')


if image_type.find('Rotated') >= 0:
    im_width = 70  # The 'RotatedPositive' has 70x70 patches and it fills image sides with black.
else: # image_type == 'Positives':
    im_width = 60  # The 'Positives' directory uses 60x60 unless the patch touches an image side.

new_im_width = im_width/2

dim = new_im_width * new_im_width
patch_size = (5, 5)
n_components = 100

# GPLVM:
GPmethod = 'BayesianGPLVM' # 'GPLVM'
LVdim = 8

showReconstructedImgs = False
showLatent = True

# Setting file names
subpatches_file = os.path.join(precompt_dir, 'subpatches_' + str(patch_size[0]) + 'x' + str(patch_size[1]) + '.pickle')
dict_file = os.path.join(precompt_dir, 'dict_' + str(dim) + 'd_' + str(n_components) + '_' + str(patch_size[0]) + 'x' + str(patch_size[1]) + '.pickle')
samples_file = os.path.join(precompt_dir, 'coded_samples_' + str(n_components) + '_' + str(patch_size[0]) + 'x' + str(patch_size[1]) + '.pickle')

gplvm_file = os.path.join(precompt_dir, GPmethod + '_' + str(dim) + 'd_' + str(n_components) + 'words_' + str(patch_size[0]) + 'x' + str(patch_size[1]) + 'patches_' + str(LVdim) + 'LVs.pickle')

do_classification_test = True

###############################################################################
if not os.path.exists(precompt_dir):
    os.mkdir(precompt_dir)


if not os.path.exists(subpatches_file):
    print('Loading all images and extracting patches.')
    im_files = os.listdir(im_dir)
    X = []

    for f in im_files:
        im = Image.open(os.path.join(im_dir,f))
        # Checking if this image includes a black border:
        im_array = np.array(im).flatten()
        if (im_array == 0).sum() < dim/10:  # only uses the image if more than 10% of it's pixels are non-zero (i.e., this is not a border)

            if (im.size[0] == im_width and im.size[1] == im_width):
                if im_width != new_im_width:
                    im = im.resize((new_im_width,new_im_width))
                
                data = extract_patches_2d(np.array(im), patch_size)
                data = data.reshape(data.shape[0], -1)

                X.append(data)

    X = np.asarray(X)
    X.dump(subpatches_file)

else:
    with file(subpatches_file) as f:
        X = cPickle.load(f)

###############################################################################
if not os.path.exists(dict_file):
    print('Preparing the data')
    Xall = np.double(X.reshape((X.shape[0]*X.shape[1],-1)))
    Xall_mean = Xall.mean(axis=0)    
    Xall -= Xall_mean
    Xall_std = Xall.std(axis=0)
    Xall /= Xall_std


    print('Learning the dictionary from reference patches.')
    dico = MiniBatchDictionaryLearning(n_components=n_components, alpha=1, n_iter=500) # This just sets the function parameters in dico
    V = dico.fit(Xall).components_ # this actually optimises and gives the dictionary    
    pb.figure(figsize=(4.2, 4))
    for i, comp in enumerate(V):
        pb.subplot(np.ceil(np.sqrt(n_components)), np.ceil(np.sqrt(n_components)), i + 1)
        pb.imshow(comp.reshape(patch_size), cmap=pb.cm.gray_r, interpolation='nearest')
        pb.xticks(())
        pb.yticks(())

    pb.suptitle('Dictionary learned from sub-patches\n', fontsize=16)
    pb.subplots_adjust(0.08, 0.02, 0.92, 0.85, 0.08, 0.23)
    with file(dict_file, 'w') as f:
        cPickle.dump((dico, Xall_mean, Xall_std), f)

else:
    with file(dict_file) as f:
        dict_data = cPickle.load(f)
        dico = dict_data[0]
        Xall_mean = dict_data[1]
        Xall_std = dict_data[2]


###############################################################################
if not os.path.exists(samples_file):
    print('Getting sparse code for all images')

    dico.set_params(transform_algorithm='omp', transform_n_nonzero_coefs=1)
    Xcoded = [];
    for i in range(X.shape[0]):
        code = dico.transform((X[i,:,:]-Xall_mean)/Xall_std)
        Xcoded.append(code)
        # Xcoded.append(code.reshape((code.shape[0]*code.shape[1],-1)).T)

    Xcoded = np.asarray(Xcoded)
    Xcoded.dump(samples_file)

else:
    with file(samples_file) as f:
        Xcoded = cPickle.load(f)   


###############################################################################
# Images have been re-coded. Now, show if their reconstructions work.

# Display the original image next to its reconstruction
def show_with_diff(source, reconstruction):
    """Helper function to display denoising"""
    # E.g.: show_with_diff(distorted, lena, 'Distorted image') # shows the source distorted image and its comparison with the original image.
    # pb.figure(figsize=(7.5, 3.3))
    pb.subplot(1, 3, 1)
    pb.title('Original')
    # pb.imshow(source, vmin=0, vmax=1, cmap=pb.cm.gray, interpolation='nearest')
    pb.imshow(source, cmap=pb.cm.gray, interpolation='nearest')
    pb.xticks(())
    pb.yticks(())
    pb.subplot(1, 3, 2)
    pb.title('Reconstruction')
    # pb.imshow(reconstruction, vmin=0, vmax=1, cmap=pb.cm.gray, interpolation='nearest')
    pb.imshow(reconstruction, cmap=pb.cm.gray, interpolation='nearest')
    pb.xticks(())
    pb.yticks(())
    pb.subplot(1, 3, 3)
    difference = source - reconstruction
    pb.title('Difference (norm: %.2f)' % np.sqrt(np.sum(difference ** 2)))
    # pb.imshow(difference, vmin=-0.5, vmax=0.5, cmap=pb.cm.PuOr, interpolation='nearest')
    pb.imshow(difference, cmap=pb.cm.PuOr, interpolation='nearest')
    pb.xticks(())
    pb.yticks(())
    # pb.suptitle('Distorted image', size=16)
    pb.subplots_adjust(0.02, 0.02, 0.98, 0.79, 0.02, 0.2)

if showReconstructedImgs:
    print('Checking reconstructions.')

    V = dico.components_
    im_files = os.listdir(im_dir)
    imIdx = 0
    pb.figure(figsize=(7.5, 3.3))
    for f in im_files:
        im = Image.open(os.path.join(im_dir,f))
        im_array = np.array(im).flatten()
        if (im_array == 0).sum() < dim/10:  # only uses the image if more than
            if (im.size[0] == im_width and im.size[1] == im_width):
                # Load original image:
                if im_width != new_im_width:
                    im = im.resize((new_im_width,new_im_width))

                # Reconstruct patches:
                patches = np.dot(Xcoded[imIdx,:,:], V)

                # Build image from patches:
                patches *= Xall_std
                patches += Xall_mean
                patches = patches.reshape(patches.shape[0], *patch_size)

                reconstruction = reconstruct_from_patches_2d(patches, (new_im_width, new_im_width))
                show_with_diff(np.asarray(im), reconstruction)
                imIdx += 1
                raw_input("Press Enter to continue...")


    # Plotting all the weights as images for the first cell in the dataset:
    pb.figure()
    for i in range(0,100):
        # codeImg = code[:,i].reshape((256-12, 128-3))
        codeImg = Xcoded[0,:,i].reshape((new_im_width-(patch_size[0]-1), new_im_width - (patch_size[1]- 1)))
        imshow(codeImg)
        raw_input("Press Enter to continue...")


    # Plotting the average weights:
    pb.figure()
    mean_code_img = Xcoded[0].reshape((new_im_width-(patch_size[0]-1), new_im_width - (patch_size[1]- 1), 100)).mean(axis=2)
    imshow(mean_code_img)

if not os.path.exists(gplvm_file):
    print('Learn GPLVM')
    Y = Xcoded.reshape(Xcoded.shape[0], -1)
    Y = Y - Y.mean()
    Y /= Y.std()
    
    if GPmethod == 'GPLVM':
        m = GPy.models.GPLVM(Y, LVdim)
        m.constrain_bounded('noise', 1e-6, 10)
    elif GPmethod == 'BayesianGPLVM':
        k = GPy.kern.rbf(LVdim,ARD=True) # + GPy.kern.bias(LVdim)# + GPy.kern.white(LVdim,0.0001)
        # k = GPy.kern.linear
        m = GPy.models.BayesianGPLVM(Y, LVdim, num_inducing=100, kernel = k)
        m['noise'] = Y.var()/100
        # m.constrain_bounded('noise', 1e-6, 10)
        m['X_variance'] *= 0.1 # 
        
    m.ensure_default_constraints()
    m.constrain_fixed('noise_variance',0.01)
    m.optimize(messages=1, max_iters=200)
    m.constrain_positive('noise_variance')
    for i in range(30):
        m.optimize(messages=1, max_iters=200)
        m.pickle(gplvm_file)

else:
    print('Loading GPLVM model from\n\t' + gplvm_file)
    with file(gplvm_file) as f:
        m = cPickle.load(f)


if showLatent:
    # Visualisation
    y = m.likelihood.Y[0,:]
    
    fig, (latent_axes, sense_axes) = pb.subplots(1, 2)
    pb.sca(latent_axes)
    
    latent_axes = m.plot_latent(ax=latent_axes)
    data_show = GPy.util.visualize.image_show(y[None,:], dimensions=(new_im_width, new_im_width), transpose=False, invert=False, scale=False, dictionary=dico.components_) # use bigX.mean() and bigX.std() here
    lvm_visualizer = GPy.util.visualize.lvm_dimselect(m.X[0, :], m, data_show, latent_axes=latent_axes, sense_axes=sense_axes)
    # raw_input('Press enter to finish')
    # pb.close('all')

m.plot_X_1d() # plots all the latent variables as a set of 1D plots -- not very helpful, as I can't tell what's the influence between one point and its neighbour.

