# Functions to generate a list of keypoints given an image
# By Teo de Campos, April 2013

import numpy as np
import cv2

import os
import Image
import pdb
import matplotlib.pyplot as plt
import collections
import SimpleCV as scv

#-------------------------------------------
def SIFT(img, patchSize=32, nPoints=500, minSize=7, maxSize=100, minColour=100): 
    # SIFT keypoint detector. Parameters:
    # * img is a numpy.array matrix
    # * nPoints is the (maximum) number of points extracted
    # * minSize is the minimum size (in pixels) of the detected feature
    # * maxSize is the maximum size (in pixels) of the detected feature
    # * patchSize is the size of features to be extracted -- though this 
    #   function does not extract features. This is only used to avoid
    #   generating patches too close to the image border.
    # * return value: array of locations pos
    detector = cv2.FeatureDetector_create('SIFT')
    oldKeyPoints = detector.detect(img)
    nKPs = len(oldKeyPoints)

    # This SIFT implementation seems to have a bug: it gives duplicates keypoints.
    # Trying to remove the repeated points:
    posID = np.zeros(nKPs)
    for i in range(0,nKPs): # Building a vector of position identifiers
        posID[i] = 10000*oldKeyPoints[i].pt[0]+oldKeyPoints[i].pt[1]

    uPosID, uPosIDidx = np.unique(posID, return_index=True)
    keyPoints = []
    for p in uPosIDidx:
        keyPoints.append(oldKeyPoints[p])

    nKPs = len(keyPoints)
    
    # Converting keypoints list to array:
    posArray = np.zeros((nKPs, 2))
    sArray = np.zeros(nKPs)
    respArray = np.zeros(nKPs)
    angleArray = np.zeros(nKPs)

    for i in range(0,nKPs):
        kp = keyPoints[i]
        respArray[i] = kp.response

    # Sorting based on response strength:    
    idx = np.argsort(-respArray)
    for i in range(0,nKPs):
        kp = keyPoints[idx[i]]
        respArray[i] = kp.response
        posArray[i] =  np.array([kp.pt[0], kp.pt[1]])
        sArray[i] = kp.size

        # converting angle to the PIL format:
        angleArray[i] = 360-kp.angle 
    
    # Remove patches that are too close to the image border:
    goodIdxFlag = (posArray[:,0]> patchSize/2) & (posArray[:,0]+patchSize/2<img.shape[1]) & (posArray[:,1]> patchSize/2) & (posArray[:,1]+patchSize/2<img.shape[0])
    if goodIdxFlag.sum() > nPoints: # prune features that are too small:
        goodIdxFlag = goodIdxFlag & (sArray>minSize)
        if goodIdxFlag.sum() > nPoints: # prune features that are too big:
            goodIdxFlag = goodIdxFlag & (sArray<maxSize)
            if goodIdxFlag.sum() > nPoints: # prune patches that are not dark enough:
                # see if the average colour of this "cell" is dark enough (the colormap is inverted):
                for i in range(0,nKPs):
                    if goodIdxFlag[i]:
                        # pdb.set_trace()
                        patch = img[posArray[i,1]-patchSize/4:posArray[i,1]+patchSize/4,posArray[i,0]-patchSize/4:posArray[i,0]+patchSize/4]
                        goodIdxFlag[i] = (patch.mean() > minColour)

    goodPos = posArray[goodIdxFlag,:]
    goodScales = sArray[goodIdxFlag,:]
    goodAngles = angleArray[goodIdxFlag,:]

    if goodIdxFlag.sum() > nPoints:
        lastIdx = nPoints
    else:
        lastIdx = goodIdxFlag.sum()

    return goodPos[0:lastIdx,:], goodScales[0:lastIdx], goodAngles[0:lastIdx]

#-------------------------------------------
def dense(img, patchSize=32, nPoints=500): 
    # Dense sampling keypoint generator. It rejects patches which are not dark enough near the centre.
    # Parameters are similar to those of SIFT.
    R = range(1+patchSize/2, img.shape[0]-patchSize/2-1, patchSize)
    C = range(1+patchSize/2, img.shape[1]-patchSize/2-1, patchSize)
    keyPoints = []
    patchMean = []
    for c in range(0,len(C)):
        for r in range(0,len(R)):
            keyPoints.append((C[c],R[r]))
            patch = img[R[r]-patchSize/4:R[r]+patchSize/4,C[c]-patchSize/4:C[c]+patchSize/4]
            patchMean.append(patch.mean())

    nKPs = len(patchMean)
    #    keyPoints = np.array(keyPoints)
    patchMean = np.array(patchMean)

    posArray = np.zeros((nKPs, 2))
    idx = np.argsort(-patchMean)
    for i in range(0,nKPs):
        posArray[i] = keyPoints[idx[i]]
    
    if nPoints ==-1:
        nPoints = nKPs

    scales = patchSize*np.ones(nPoints)
    angles = np.zeros(nPoints)
    return posArray[0:nPoints,:], scales, angles
#-------------------------------------------
def Oxford(mask):
    img = scv.Image(mask)
    blobs = img.findBlobs() # this gives blobs sorted by increasing area 
    nKPs = len(blobs)
    posArray = np.zeros((nKPs, 2))
    scaleArray = np.zeros(nKPs)
    angleArray = np.zeros(nKPs)

    for i in range(nKPs):
        posArray[i,:] = np.array((blobs[i].y,blobs[i].x))
        scaleArray[i] = blobs[i].radius()
        angleArray[i] = blobs[i].angle()
        
    return posArray, scaleArray, angleArray

#-------------------------------------------
def plotKeyPoints(img, KPs, scales):
    plt.ion() 
    fig = plt.figure(frameon=False)
    plt.imshow(255-img,cmap='gray')
    for i in range(0,KPs.shape[0]):
        plt.plot(KPs[i,0],KPs[i,1],'y*')
        circle=plt.Circle((KPs[i,0],KPs[i,1]),scales[i]/2,color='r',fill=False)
        fig.gca().add_artist(circle)

    frame1 = plt.gca()
    frame1.axes.get_xaxis().set_visible(False)
    frame1.axes.get_yaxis().set_visible(False)
    plt.show()

#-------------------------------------------
if __name__ == '__main__':
    # Test function
    datadir = os.path.join(os.path.expanduser('~'), 'Datasets/StanfordTissue', 'NKI_HE', 'NKI_Images')
    imageFiles = os.listdir(datadir)

    im = Image.open(os.path.join(datadir, imageFiles[0])).quantize() 
    img = np.array(im)
    
    KPs, scales = SIFT(img,patchSize=32, nPoints=500,minSize=7, maxSize=64, minColour=125)
    # KPs, scales = dense(img,patchSize=32, nPoints=500)
    print str(KPs.shape[0]) + ' keypoints'
    plotKeyPoints(img, KPs, scales)
